"""HTTP client for the Adstract AI SDK."""

from __future__ import annotations

import asyncio
import hashlib
import json
import logging
import os
import time
from importlib import metadata as importlib_metadata
from typing import Any

import httpx

from adstractai.constants import (
    AD_INJECTION_ENDPOINT,
    API_KEY_HEADER_NAME,
    BASE_URL,
    DEFAULT_RETRIES,
    DEFAULT_TIMEOUT_SECONDS,
    ENV_API_KEY_NAME,
    MAX_RETRIES,
    SDK_HEADER_NAME,
    SDK_NAME,
    SDK_VERSION,
    SDK_VERSION_HEADER_NAME,
)
from adstractai.errors import (
    AdEnhancementError,
    AdSDKError,
    AuthenticationError,
    MissingParameterError,
    NetworkError,
    RateLimitError,
    ServerError,
    UnexpectedResponseError,
    ValidationError,
)
from adstractai.models import AdRequest, AdResponse, Constraints, Conversation, Metadata

logger = logging.getLogger(__name__)

MIN_API_KEY_LENGTH = 10
MIN_USER_AGENT_LENGTH = 10
RATE_LIMIT_STATUS = 429
SERVER_ERROR_MIN = 500
SERVER_ERROR_MAX = 599
CLIENT_ERROR_MIN = 400
CLIENT_ERROR_MAX = 499


class Adstract:
    """Client for sending ad requests to the Adstract backend."""

    def __init__(
        self,
        *,
        api_key: str | None = None,
        base_url: str | None = None,
        timeout: float = DEFAULT_TIMEOUT_SECONDS,
        retries: int = DEFAULT_RETRIES,
        backoff_factor: float = 0.5,
        max_backoff: float = 8.0,
        http_client: httpx.Client | None = None,
        async_http_client: httpx.AsyncClient | None = None,
    ) -> None:
        if api_key is None:
            api_key = os.environ.get(ENV_API_KEY_NAME)
        if not isinstance(api_key, str) or len(api_key.strip()) < MIN_API_KEY_LENGTH:
            raise ValidationError("api_key must be at least 10 characters")
        self._api_key = api_key
        self._base_url = base_url or BASE_URL
        self._timeout = timeout
        self._retries = retries if retries <= MAX_RETRIES else DEFAULT_RETRIES
        self._backoff_factor = backoff_factor
        self._max_backoff = max_backoff
        self._client = http_client or httpx.Client(timeout=timeout)
        self._async_client = async_http_client or httpx.AsyncClient(timeout=timeout)
        self._owns_client = http_client is None
        self._owns_async_client = async_http_client is None

    def close(self) -> None:
        if self._owns_client:
            self._client.close()

    async def aclose(self) -> None:
        if self._owns_async_client:
            await self._async_client.aclose()

    def request_ad_enhancement(
        self,
        *,
        prompt: str,
        conversation: dict[str, Any] | Conversation,
        user_agent: str,
        x_forwarded_for: str,
        constraints: dict[str, Any] | Constraints | None = None,
    ) -> str:
        if not user_agent:
            raise MissingParameterError("user_agent parameter is required")
        if not x_forwarded_for:
            raise MissingParameterError("x_forwarded_for parameter is required")

        metadata = self._build_metadata(
            user_agent=user_agent,
            x_forwarded_for=x_forwarded_for,
        )
        request_model = AdRequest.from_values(
            prompt=prompt,
            conversation=conversation,
            metadata=metadata,
            constraints=constraints,
        )
        payload = request_model.to_payload()
        logger.debug(
            "Sending ad enhancement request", extra={"prompt_length": len(request_model.prompt)}
        )

        response = self._send_request(payload)

        # Check if enhancement was successful
        if not response.success:
            raise AdEnhancementError(
                "Ad enhancement failed",
                status_code=None,
                response_snippet=f"success: {response.success}",
            )

        # Check if aepi data is available
        if response.aepi is None or response.aepi.aepi_text is None:
            raise AdEnhancementError(
                "Ad enhancement response missing aepi data",
                status_code=None,
                response_snippet="aepi or aepi_text is None",
            )

        return response.aepi.aepi_text

    def request_ad_enhancement_or_default(
        self,
        *,
        prompt: str,
        conversation: dict[str, Any] | Conversation,
        user_agent: str,
        x_forwarded_for: str,
        constraints: dict[str, Any] | Constraints | None = None,
    ) -> str:
        if not user_agent:
            raise MissingParameterError("user_agent parameter is required")
        if not x_forwarded_for:
            raise MissingParameterError("x_forwarded_for parameter is required")

        try:
            metadata = self._build_metadata(
                user_agent=user_agent,
                x_forwarded_for=x_forwarded_for,
            )
            request_model = AdRequest.from_values(
                prompt=prompt,
                conversation=conversation,
                metadata=metadata,
                constraints=constraints,
            )
            payload = request_model.to_payload()
            logger.debug(
                "Sending ad enhancement request (with fallback)",
                extra={"prompt_length": len(request_model.prompt)},
            )

            response = self._send_request(payload)

            # Check if enhancement was successful and has aepi data
            if (
                response.success
                and response.aepi is not None
                and response.aepi.aepi_text is not None
            ):
                return response.aepi.aepi_text
            else:
                logger.debug(
                    "Enhancement not successful or missing aepi data, returning original prompt"
                )
                return prompt

        except Exception as exc:
            logger.debug(
                "Enhancement failed with exception, returning original prompt", exc_info=exc
            )
            return prompt

    async def request_ad_enhancement_async(
        self,
        *,
        prompt: str,
        conversation: dict[str, Any] | Conversation,
        user_agent: str,
        x_forwarded_for: str,
        constraints: dict[str, Any] | Constraints | None = None,
    ) -> str:
        if not user_agent:
            raise MissingParameterError("user_agent parameter is required")
        if not x_forwarded_for:
            raise MissingParameterError("x_forwarded_for parameter is required")

        metadata = self._build_metadata(
            user_agent=user_agent,
            x_forwarded_for=x_forwarded_for,
        )
        request_model = AdRequest.from_values(
            prompt=prompt,
            conversation=conversation,
            metadata=metadata,
            constraints=constraints,
        )
        payload = request_model.to_payload()
        logger.debug(
            "Sending async ad enhancement request",
            extra={"prompt_length": len(request_model.prompt)},
        )

        response = await self._send_request_async(payload)

        # Check if enhancement was successful
        if not response.success:
            raise AdEnhancementError(
                "Ad enhancement failed",
                status_code=None,
                response_snippet=f"success: {response.success}",
            )

        # Check if aepi data is available
        if response.aepi is None or response.aepi.aepi_text is None:
            raise AdEnhancementError(
                "Ad enhancement response missing aepi data",
                status_code=None,
                response_snippet="aepi or aepi_text is None",
            )

        return response.aepi.aepi_text

    async def request_ad_enhancement_or_default_async(
        self,
        *,
        prompt: str,
        conversation: dict[str, Any] | Conversation,
        user_agent: str,
        x_forwarded_for: str,
        constraints: dict[str, Any] | Constraints | None = None,
    ) -> str:
        if not user_agent:
            raise MissingParameterError("user_agent parameter is required")
        if not x_forwarded_for:
            raise MissingParameterError("x_forwarded_for parameter is required")

        try:
            metadata = self._build_metadata(
                user_agent=user_agent,
                x_forwarded_for=x_forwarded_for,
            )
            request_model = AdRequest.from_values(
                prompt=prompt,
                conversation=conversation,
                metadata=metadata,
                constraints=constraints,
            )
            payload = request_model.to_payload()
            logger.debug(
                "Sending async ad enhancement request (with fallback)",
                extra={"prompt_length": len(request_model.prompt)},
            )

            response = await self._send_request_async(payload)

            # Check if enhancement was successful and has aepi data
            if (
                response.success
                and response.aepi is not None
                and response.aepi.aepi_text is not None
            ):
                return response.aepi.aepi_text
            else:
                logger.debug(
                    "Enhancement not successful or missing aepi data, returning original prompt"
                )
                return prompt

        except Exception as exc:
            logger.debug(
                "Enhancement failed with exception, returning original prompt", exc_info=exc
            )
            return prompt

    def _endpoint(self) -> str:
        return f"{self._base_url}{AD_INJECTION_ENDPOINT}"

    def _send_request(self, payload: dict[str, Any]) -> AdResponse:
        url = self._endpoint()
        headers = self._build_headers()
        for attempt in range(self._retries + 1):
            try:
                response = self._client.post(
                    url, json=payload, headers=headers, timeout=self._timeout
                )
            except (httpx.TimeoutException, httpx.TransportError) as exc:
                logger.debug("Network error on attempt %s", attempt + 1)
                if attempt >= self._retries:
                    raise NetworkError("Network error during request", original_error=exc) from exc
                self._sleep_backoff(attempt)
                continue

            if response.status_code == RATE_LIMIT_STATUS:
                logger.debug("Rate limited on attempt %s", attempt + 1)
                if attempt >= self._retries:
                    raise RateLimitError(
                        "Rate limited",
                        status_code=response.status_code,
                        response_snippet=_snippet(response),
                    )
                self._sleep_backoff(attempt)
                continue
            if SERVER_ERROR_MIN <= response.status_code <= SERVER_ERROR_MAX:
                logger.debug("Server error on attempt %s", attempt + 1)
                if attempt >= self._retries:
                    raise ServerError(
                        "Server error",
                        status_code=response.status_code,
                        response_snippet=_snippet(response),
                    )
                self._sleep_backoff(attempt)
                continue

            return self._handle_response(response)

        raise AdSDKError("Unhandled retry loop exit")

    async def _send_request_async(self, payload: dict[str, Any]) -> AdResponse:
        url = self._endpoint()
        headers = self._build_headers()
        for attempt in range(self._retries + 1):
            try:
                response = await self._async_client.post(
                    url, json=payload, headers=headers, timeout=self._timeout
                )
            except (httpx.TimeoutException, httpx.TransportError) as exc:
                logger.debug("Async network error on attempt %s", attempt + 1)
                if attempt >= self._retries:
                    raise NetworkError("Network error during request", original_error=exc) from exc
                await self._sleep_backoff_async(attempt)
                continue

            if response.status_code == RATE_LIMIT_STATUS:
                logger.debug("Async rate limited on attempt %s", attempt + 1)
                if attempt >= self._retries:
                    raise RateLimitError(
                        "Rate limited",
                        status_code=response.status_code,
                        response_snippet=_snippet(response),
                    )
                await self._sleep_backoff_async(attempt)
                continue
            if SERVER_ERROR_MIN <= response.status_code <= SERVER_ERROR_MAX:
                logger.debug("Async server error on attempt %s", attempt + 1)
                if attempt >= self._retries:
                    raise ServerError(
                        "Server error",
                        status_code=response.status_code,
                        response_snippet=_snippet(response),
                    )
                await self._sleep_backoff_async(attempt)
                continue

            return self._handle_response(response)

        raise AdSDKError("Unhandled retry loop exit")

    def _handle_response(self, response: httpx.Response) -> AdResponse:
        status = response.status_code
        if status in {401, 403}:
            raise AuthenticationError(
                "Authentication failed",
                status_code=status,
                response_snippet=_snippet(response),
            )
        if CLIENT_ERROR_MIN <= status <= CLIENT_ERROR_MAX:
            raise UnexpectedResponseError(
                "Unexpected client error",
                status_code=status,
                response_snippet=_snippet(response),
            )

        try:
            data = response.json()
        except json.JSONDecodeError as exc:
            raise UnexpectedResponseError(
                "Invalid JSON response",
                status_code=status,
                response_snippet=_snippet(response),
            ) from exc

        try:
            return AdResponse.from_json(data)
        except ValidationError as exc:
            raise UnexpectedResponseError(
                "Unexpected response structure",
                status_code=status,
                response_snippet=_snippet(response),
            ) from exc

    def _sleep_backoff(self, attempt: int) -> None:
        delay = min(self._backoff_factor * (2**attempt), self._max_backoff)
        time.sleep(delay)

    async def _sleep_backoff_async(self, attempt: int) -> None:
        delay = min(self._backoff_factor * (2**attempt), self._max_backoff)
        await asyncio.sleep(delay)

    def _build_headers(self) -> dict[str, str]:
        return {
            SDK_HEADER_NAME: SDK_NAME,
            SDK_VERSION_HEADER_NAME: SDK_VERSION,
            API_KEY_HEADER_NAME: self._api_key,
        }

    def _build_metadata(
        self,
        *,
        user_agent: str,
        x_forwarded_for: str,
    ) -> Metadata:
        if len(user_agent) < MIN_USER_AGENT_LENGTH:
            raise ValidationError("user_agent is invalid")

        derived_client = _build_client_metadata(user_agent, x_forwarded_for)
        metadata_dict = {"client": derived_client}

        try:
            return Metadata.model_validate(metadata_dict)
        except Exception as exc:
            raise ValidationError("Failed to build metadata") from exc


def _snippet(response: httpx.Response, limit: int = 200) -> str | None:
    if response.text is None:
        return None
    return response.text[:limit]


def _build_client_metadata(user_agent: str, x_forwarded_for: str) -> dict[str, Any]:
    user_agent_hash = hashlib.sha256(user_agent.encode("utf-8")).hexdigest()
    os_family = _parse_os_family(user_agent)
    browser_family = _parse_browser_family(user_agent)
    device_type = _parse_device_type(user_agent)
    sdk_version = _sdk_version()
    client: dict[str, Any] = {
        "user_agent_hash": user_agent_hash,
        "device_type": device_type,
        "sdk_version": sdk_version,
        "x_forwarded_for": x_forwarded_for,
    }
    if os_family:
        client["os_family"] = os_family
    if browser_family:
        client["browser_family"] = browser_family
    return client


def _parse_device_type(user_agent: str) -> str:
    value = user_agent.lower()
    if any(token in value for token in ["bot", "crawler", "spider", "slurp", "bingpreview"]):
        return "bot"
    if "ipad" in value or "tablet" in value:
        return "tablet"
    if "mobile" in value or "iphone" in value or "android" in value:
        return "mobile"
    if any(token in value for token in ["windows", "macintosh", "linux", "cros"]):
        return "desktop"
    return "unknown"


def _parse_os_family(user_agent: str) -> str | None:
    value = user_agent.lower()
    candidates = (
        ("windows", "Windows"),
        ("android", "Android"),
        ("iphone", "iOS"),
        ("ipad", "iOS"),
        ("ios", "iOS"),
        ("mac os x", "macOS"),
        ("macintosh", "macOS"),
        ("cros", "ChromeOS"),
        ("linux", "Linux"),
    )
    for token, label in candidates:
        if token in value:
            return label
    return None


def _parse_browser_family(user_agent: str) -> str | None:
    value = user_agent.lower()
    if "edg" in value:
        result = "Edge"
    elif "opr" in value or "opera" in value:
        result = "Opera"
    elif "chrome" in value and "chromium" not in value and "edg" not in value:
        result = "Chrome"
    elif "safari" in value and "chrome" not in value and "chromium" not in value:
        result = "Safari"
    elif "firefox" in value:
        result = "Firefox"
    elif "chromium" in value:
        result = "Chromium"
    else:
        result = None
    return result


def _sdk_version() -> str:
    try:
        return importlib_metadata.version("adstractai")
    except importlib_metadata.PackageNotFoundError:
        return "0.0.0"



"""Typed request/response models for the Adstract AI SDK."""

from __future__ import annotations

import re
from typing import Any

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator
from pydantic import ValidationError as PydanticValidationError

from adstractai.errors import ValidationError

_SEMVER_RE = re.compile(r"^\d+\.\d+\.\d+$")


class Conversation(BaseModel):
    model_config = ConfigDict(extra="forbid")

    conversation_id: str = Field(min_length=1)
    session_id: str = Field(min_length=1)
    message_id: str = Field(min_length=1)


class ClientMetadata(BaseModel):
    model_config = ConfigDict(extra="forbid")

    ip_hash: str | None = None
    os_family: str | None = None
    device_type: str | None = None
    referrer: str | None = None
    x_forwarded_for: str | None = None
    user_agent_hash: str | None = None
    browser_family: str | None = None
    sdk_version: str | None = None

    @field_validator("sdk_version")
    @classmethod
    def _validate_sdk_version(cls, value: str | None) -> str | None:
        if value is None:
            return None
        if not _SEMVER_RE.match(value):
            raise ValueError("sdk_version must be a semver string")
        return value


class Constraints(BaseModel):
    model_config = ConfigDict(extra="forbid")

    max_ads: int = Field(default=1, ge=1, le=20)
    min_similarity_hint: float | None = Field(default=None, ge=0.0, le=1.0)
    max_latency_ms_hint: int | None = Field(default=None, ge=0)
    safe_mode: str = "standard"

    @field_validator("safe_mode")
    @classmethod
    def _validate_safe_mode(cls, value: str) -> str:
        if value not in {"strict", "standard", "off"}:
            raise ValueError("safe_mode must be 'strict', 'standard', or 'off'")
        return value


class Metadata(BaseModel):
    model_config = ConfigDict(extra="forbid")

    client: ClientMetadata | None = None

    @model_validator(mode="after")
    def _ensure_client(self) -> Metadata:
        if self.client is None:
            raise ValueError("metadata must include client")
        return self


class AdRequest(BaseModel):
    model_config = ConfigDict(extra="forbid")

    prompt: str = Field(min_length=3)
    conversation: Conversation
    metadata: Metadata | None = None
    constraints: Constraints | None = None

    @classmethod
    def from_values(
        cls,
        *,
        prompt: Any,
        conversation: Any,
        metadata: Any = None,
        constraints: Any = None,
    ) -> AdRequest:
        try:
            return cls.model_validate(
                {
                    "prompt": prompt,
                    "conversation": conversation,
                    "metadata": metadata,
                    "constraints": constraints,
                }
            )
        except PydanticValidationError as exc:
            raise ValidationError("Invalid request payload") from exc

    def to_payload(self) -> dict[str, Any]:
        return self.model_dump(exclude_none=True)


class AepiData(BaseModel):
    model_config = ConfigDict(extra="allow")

    status: str
    aepi_text: str
    checksum: str
    size_bytes: int


class AdResponse(BaseModel):
    model_config = ConfigDict(extra="allow")

    raw: dict[str, Any]
    ad_request_id: str | None = None
    ad_response_id: str | None = None
    success: bool | None = None
    execution_time_ms: float | None = None
    aepi: AepiData | None = None
    tracking_url: str | None = None
    product_name: str | None = None

    @classmethod
    def from_json(cls, payload: Any) -> AdResponse:
        if not isinstance(payload, dict):
            raise ValidationError("response JSON must be an object")

        # Parse aepi data if present
        aepi = None
        aepi_data = payload.get("aepi")
        if aepi_data is not None:
            if not isinstance(aepi_data, dict):
                raise ValidationError("aepi must be an object")
            try:
                aepi = AepiData.model_validate(aepi_data)
            except PydanticValidationError as exc:
                raise ValidationError("Invalid aepi data structure") from exc

        try:
            return cls.model_validate(
                {
                    "raw": payload,
                    "ad_request_id": payload.get("ad_request_id"),
                    "ad_response_id": payload.get("ad_response_id"),
                    "success": payload.get("success"),
                    "execution_time_ms": payload.get("execution_time_ms"),
                    "aepi": aepi,
                    "tracking_url": payload.get("tracking_url"),
                    "product_name": payload.get("product_name"),
                }
            )
        except PydanticValidationError as exc:
            raise ValidationError("response JSON validation failed") from exc



import asyncio
import json
from typing import TypedDict

import httpx
import pytest

from adstractai.client import Adstract
from adstractai.constants import (
    API_KEY_HEADER_NAME,
    SDK_HEADER_NAME,
    SDK_NAME,
    SDK_VERSION,
    SDK_VERSION_HEADER_NAME,
)
from adstractai.errors import (
    AuthenticationError,
    MissingParameterError,
    RateLimitError,
    ServerError,
)

DEFAULT_USER_AGENT = (
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 "
    "(KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36"
)
RETRY_SUCCESS_AFTER = 1

API_KEY = "adpk_live_gx6xbutnrkyjaqjd.uatnQaAhIho-QalyI5Cng3CRhJKobYWoBGFqrvzgdPQ"
X_FORWARDED_FOR = "185.100.245.160"


class ValidPayload(TypedDict):
    prompt: str
    conversation: dict[str, str]
    user_agent: str


def _valid_payload(user_agent: str = DEFAULT_USER_AGENT) -> ValidPayload:
    return {
        "prompt": "Explain ad targeting",
        "conversation": {
            "conversation_id": "conv-1",
            "session_id": "sess-1",
            "message_id": "msg-1",
        },
        "user_agent": user_agent,
    }


def test_headers_include_sdk_and_api_key() -> None:
    captured = {}

    def handler(request: httpx.Request) -> httpx.Response:
        captured["headers"] = request.headers
        return httpx.Response(
            200,
            json={
                "ad_request_id": "test-3",
                "ad_response_id": "test-3",
                "success": True,
                "execution_time_ms": 100.0,
            },
        )

    transport = httpx.MockTransport(handler)
    client = Adstract(
        api_key=API_KEY,
        http_client=httpx.Client(transport=transport),
    )

    payload = _valid_payload()
    client.request_ad_enhancement_or_default(**payload, x_forwarded_for=X_FORWARDED_FOR)

    headers = captured["headers"]
    assert headers[SDK_HEADER_NAME] == SDK_NAME
    assert headers[SDK_VERSION_HEADER_NAME] == SDK_VERSION
    assert headers[API_KEY_HEADER_NAME] == API_KEY


def test_client_metadata_generated_from_user_agent() -> None:
    captured = {}

    def handler(request: httpx.Request) -> httpx.Response:
        captured["payload"] = json.loads(request.content.decode("utf-8"))
        return httpx.Response(
            200,
            json={
                "ad_request_id": "test-4",
                "ad_response_id": "test-4",
                "success": True,
                "execution_time_ms": 100.0,
            },
        )

    transport = httpx.MockTransport(handler)
    client = Adstract(
        api_key=API_KEY,
        http_client=httpx.Client(transport=transport),
    )

    payload = _valid_payload(
        user_agent=(
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 "
            "(KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36"
        )
    )
    client.request_ad_enhancement_or_default(**payload, x_forwarded_for="8.8.8.8")

    sent = captured["payload"]
    assert "metadata" in sent
    assert "client" in sent["metadata"]
    assert "user_agent_hash" in sent["metadata"]["client"]
    assert sent["metadata"]["client"]["device_type"] in {
        "desktop",
        "mobile",
        "tablet",
        "bot",
        "unknown",
    }
    assert "geo" not in sent["metadata"]


def test_geo_not_generated_without_provider() -> None:
    captured = {}

    def handler(request: httpx.Request) -> httpx.Response:
        captured["payload"] = json.loads(request.content.decode("utf-8"))
        return httpx.Response(
            200,
            json={
                "ad_request_id": "test-5",
                "ad_response_id": "test-5",
                "success": True,
                "execution_time_ms": 100.0,
            },
        )

    transport = httpx.MockTransport(handler)
    client = Adstract(
        api_key=API_KEY,
        http_client=httpx.Client(transport=transport),
    )

    payload = _valid_payload(
        user_agent=(
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 "
            "(KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36"
        )
    )
    client.request_ad_enhancement_or_default(
        **payload,
        x_forwarded_for="8.8.4.4",
    )

    sent = captured["payload"]
    assert "metadata" in sent
    assert "client" in sent["metadata"]
    assert "geo" not in sent["metadata"]


def test_geo_generated_with_provider() -> None:
    captured = {}
    resolved: dict[str, str | None] = {"ip": None}

    def handler(request: httpx.Request) -> httpx.Response:
        captured["payload"] = json.loads(request.content.decode("utf-8"))
        return httpx.Response(
            200,
            json={
                "ad_request_id": "test-6",
                "ad_response_id": "test-6",
                "success": True,
                "execution_time_ms": 100.0,
            },
        )

    def geo_provider(ip: str) -> dict[str, object]:
        resolved["ip"] = ip
        return {"geo_country": "US", "geo_region": "CA"}

    transport = httpx.MockTransport(handler)
    client = Adstract(
        api_key=API_KEY,
        http_client=httpx.Client(transport=transport),
    )

    payload = _valid_payload(
        user_agent=(
            "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 "
            "(KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36"
        )
    )
    client.request_ad_enhancement_or_default(
        **payload,
        x_forwarded_for="10.0.0.1, 8.8.8.8",
    )

    sent = captured["payload"]
    assert resolved["ip"] is None  # geo_provider not called since not supported
    assert "geo" not in sent["metadata"]  # No geo metadata without geo support


@pytest.mark.parametrize("status", [401, 403])
def test_authentication_error_mapping(status: int) -> None:
    def handler(request: httpx.Request) -> httpx.Response:
        return httpx.Response(status, json={"detail": "denied"})

    transport = httpx.MockTransport(handler)
    client = Adstract(
        api_key=API_KEY,
        http_client=httpx.Client(transport=transport),
    )

    with pytest.raises(AuthenticationError):
        client.request_ad_enhancement(**_valid_payload(), x_forwarded_for=X_FORWARDED_FOR)


def test_rate_limit_mapping() -> None:
    def handler(request: httpx.Request) -> httpx.Response:
        return httpx.Response(429, json={"detail": "slow down"})

    transport = httpx.MockTransport(handler)
    client = Adstract(
        api_key=API_KEY,
        http_client=httpx.Client(transport=transport),
        retries=0,
    )

    with pytest.raises(RateLimitError):
        client.request_ad_enhancement(**_valid_payload(), x_forwarded_for=X_FORWARDED_FOR)


def test_server_error_mapping() -> None:
    def handler(request: httpx.Request) -> httpx.Response:
        return httpx.Response(500, json={"detail": "error"})

    transport = httpx.MockTransport(handler)
    client = Adstract(
        api_key=API_KEY,
        http_client=httpx.Client(transport=transport),
        retries=0,
    )

    with pytest.raises(ServerError):
        client.request_ad_enhancement(**_valid_payload(), x_forwarded_for=X_FORWARDED_FOR)


def test_retry_then_success() -> None:
    calls = {"count": 0}

    def handler(request: httpx.Request) -> httpx.Response:
        calls["count"] += 1
        if calls["count"] < RETRY_SUCCESS_AFTER:
            return httpx.Response(500, json={"detail": "error"})
        return httpx.Response(
            200,
            json={
                "ad_request_id": "test-req-id",
                "ad_response_id": "test-resp-id",
                "success": True,
                "execution_time_ms": 100.5,
                "aepi": {
                    "status": "ok",
                    "aepi_text": "Test ad content",
                    "checksum": "test-checksum",
                    "size_bytes": 100,
                },
                "tracking_url": "http://example.com/track",
                "product_name": "Test Product",
            },
        )

    transport = httpx.MockTransport(handler)
    client = Adstract(
        api_key=API_KEY,
        http_client=httpx.Client(transport=transport),
        retries=1,
        backoff_factor=0.0,
    )

    aepi_text = client.request_ad_enhancement(**_valid_payload(), x_forwarded_for=X_FORWARDED_FOR)

    assert calls["count"] == RETRY_SUCCESS_AFTER
    assert aepi_text == "Test ad content"


def test_async_request() -> None:
    async def run_test() -> None:
        def handler(request: httpx.Request) -> httpx.Response:
            return httpx.Response(
                200,
                json={
                    "ad_request_id": "async-req-id",
                    "ad_response_id": "async-resp-id",
                    "success": True,
                    "execution_time_ms": 200.0,
                    "aepi": {
                        "status": "ok",
                        "aepi_text": "Async test ad content",
                        "checksum": "async-test-checksum",
                        "size_bytes": 150,
                    },
                    "tracking_url": "http://example.com/async-track",
                    "product_name": "Async Test Product",
                },
            )

        transport = httpx.MockTransport(handler)
        async_client = httpx.AsyncClient(transport=transport)
        client = Adstract(api_key=API_KEY, async_http_client=async_client)

        aepi_text = await client.request_ad_enhancement_async(
            **_valid_payload(), x_forwarded_for=X_FORWARDED_FOR
        )
        assert aepi_text == "Async test ad content"

        await client.aclose()

    asyncio.run(run_test())


def test_new_response_format() -> None:
    """Test that the new response format with aepi data is properly parsed."""
    new_format_response = {
        "ad_request_id": "ac12d9db-e7f8-42f2-a101-7eed89693c43",
        "ad_response_id": "ac12d9db-e7f8-42f2-a101-7eed89693c43",
        "success": True,
        "execution_time_ms": 1025.6521701812744,
        "aepi": {
            "status": "ok",
            "aepi_text": "You are an AI assistant that integrates advertisements...",
            "checksum": "3683054a04bcb31e186e9439c6d2d0b0fd36b70c8d53c5ab3e847402418fb688",
            "size_bytes": 1726,
        },
        "tracking_url": "http://localhost:8000/c/2uuRF2WHizzzYedtMcShvv",
        "product_name": "Adstract – LLM Advertising",
    }

    def handler(request: httpx.Request) -> httpx.Response:
        return httpx.Response(200, json=new_format_response)

    transport = httpx.MockTransport(handler)
    client = Adstract(
        api_key=API_KEY,
        http_client=httpx.Client(transport=transport),
    )

    aepi_text = client.request_ad_enhancement(**_valid_payload(), x_forwarded_for=X_FORWARDED_FOR)

    # Test that the aepi_text is correctly extracted from the new response format
    assert aepi_text == "You are an AI assistant that integrates advertisements..."


def test_missing_user_agent_raises_exception() -> None:
    """Test that missing user_agent parameter raises MissingParameterError."""
    client = Adstract(api_key=API_KEY)

    with pytest.raises(MissingParameterError, match="user_agent parameter is required"):
        client.request_ad_enhancement(
            prompt="Test prompt",
            conversation={"conversation_id": "c", "session_id": "s", "message_id": "m"},
            user_agent="",  # Empty string should trigger the error
            x_forwarded_for=X_FORWARDED_FOR,
        )


def test_missing_x_forwarded_for_raises_exception() -> None:
    """Test that missing x_forwarded_for parameter raises MissingParameterError."""
    client = Adstract(api_key=API_KEY)

    with pytest.raises(MissingParameterError, match="x_forwarded_for parameter is required"):
        client.request_ad_enhancement(
            prompt="Test prompt",
            conversation={"conversation_id": "c", "session_id": "s", "message_id": "m"},
            user_agent=DEFAULT_USER_AGENT,
            x_forwarded_for="",  # Empty string should trigger the error
        )


def test_missing_parameters_in_or_default_method() -> None:
    """Test that missing parameters raise MissingParameterError in or_default method."""
    client = Adstract(api_key=API_KEY)

    with pytest.raises(MissingParameterError, match="user_agent parameter is required"):
        client.request_ad_enhancement_or_default(
            prompt="Test prompt",
            conversation={"conversation_id": "c", "session_id": "s", "message_id": "m"},
            user_agent="",
            x_forwarded_for=X_FORWARDED_FOR,
        )

    with pytest.raises(MissingParameterError, match="x_forwarded_for parameter is required"):
        client.request_ad_enhancement_or_default(
            prompt="Test prompt",
            conversation={"conversation_id": "c", "session_id": "s", "message_id": "m"},
            user_agent=DEFAULT_USER_AGENT,
            x_forwarded_for="",
        )


def test_missing_parameters_in_async_methods() -> None:
    """Test that missing parameters raise MissingParameterError in async methods."""

    async def run_test() -> None:
        client = Adstract(api_key=API_KEY)

        with pytest.raises(MissingParameterError, match="user_agent parameter is required"):
            await client.request_ad_enhancement_async(
                prompt="Test prompt",
                conversation={"conversation_id": "c", "session_id": "s", "message_id": "m"},
                user_agent="",
                x_forwarded_for=X_FORWARDED_FOR,
            )

        with pytest.raises(MissingParameterError, match="x_forwarded_for parameter is required"):
            await client.request_ad_enhancement_async(
                prompt="Test prompt",
                conversation={"conversation_id": "c", "session_id": "s", "message_id": "m"},
                user_agent=DEFAULT_USER_AGENT,
                x_forwarded_for="",
            )

        with pytest.raises(MissingParameterError, match="user_agent parameter is required"):
            await client.request_ad_enhancement_or_default_async(
                prompt="Test prompt",
                conversation={"conversation_id": "c", "session_id": "s", "message_id": "m"},
                user_agent="",
                x_forwarded_for=X_FORWARDED_FOR,
            )

        with pytest.raises(MissingParameterError, match="x_forwarded_for parameter is required"):
            await client.request_ad_enhancement_or_default_async(
                prompt="Test prompt",
                conversation={"conversation_id": "c", "session_id": "s", "message_id": "m"},
                user_agent=DEFAULT_USER_AGENT,
                x_forwarded_for="",
            )

        await client.aclose()

    asyncio.run(run_test())



"""Constants for the Adstract AI SDK."""

import os

BASE_URL = os.getenv("ADSTRACT_DEBUG_URL", "https://api.adstract.ai")
AD_INJECTION_ENDPOINT = "/api/ad-injection/start/"

DEFAULT_TIMEOUT_SECONDS = 100

DEFAULT_RETRIES = 0
MAX_RETRIES = 1

ENV_API_KEY_NAME = "ADSTRACT_API_KEY"

SDK_HEADER_NAME = "X-Adstract-SDK"
SDK_VERSION_HEADER_NAME = "X-Adstract-SDK-Version"
API_KEY_HEADER_NAME = "X-Adstract-API-Key"

SDK_NAME = "adstractai-python"
SDK_VERSION = "0.0.9"
